/* Copyright(c) 1994  Danny Sadinoff
  See section COPYING for conditions for redistribution
*/
#ifndef __SEDRA__
#define __SEDRA__

int sedra( long int absDate, char *out_buf, int buf_len );
void reset_sedra( int );

#endif
